# Flex processor

Flex Process is a versatile Drupal module designed to empower developers with
the ability to attach customizable processor plugins to various data structures.
This tool allows developers to define specific processing workflows for their
data, enabling dynamic manipulation and transformation.

For a full description of the module, visit the
[project page](https://www.drupal.org/project/flex_processor).

Submit bug reports and feature suggestions, or track changes in the
[issue queue](https://www.drupal.org/project/issues/flex_processor).


## Requirements

This module requires no modules outside of Drupal core.


## Installation

Install as you would normally install a contributed Drupal module. For further
information, see
[Installing Drupal Modules](https://www.drupal.org/docs/extending-drupal/installing-drupal-modules).


## Configuration

The module has no menu or modifiable settings. There is no configuration. When
enabled, the module will provide the helpers to create the plugins.


## How it works

This module provides the ability to create data processor plugins in a simple
way. Every time a data, form now on data component, is sent to be processed, the
corresponding plugin matching that data component type is executed. The
processors plugin are defined in a flexible way that any kind of processing task
is allowed. How to process to process the dataComponent is up to the developer,
and they need.


## Defining a Processor plugin

There are 3 predefined processor types: Entities, Fields, DataComponents; But
any other type can be registered.


### Entity processor

Entities are one of the default known data structures and can be processed with
no effort.

**Here is an example of a processor used to return a formatted structure for an**
**Article node bundle**

> file: modules/custom/your_module/src/Plugin/EntityDataProcessor/ArticleCard.php
```php
<?php
namespace Drupal\your_module\Plugin\FieldDataProcessor;

use Drupal\flex_processor\Plugin\EntityDataProcessor;

/**
 * Returns the structured data of an entity.
 *
 * @DataProcessor(
 *   id = "node__article__card",
 *   label = @Translation("Node: Article Card"),
 *   type = "node",
 *   bundles = {
 *     "article"
 *   },
 *   variant = "card"
 * )
 */
class ArticleCard extends EntityDataProcessor {

  /**
   * {@inheritdoc}
   */
  public function process(mixed $entity, array $options = []): mixed {

    return [
      'title' => $entity->getTitle(),
      'type' => $entity->getType(),
      'url' => $entity->toUrl()->toString(TRUE)->getGeneratedUrl(),
      'fields' => [

        // Triggers processor in the article body field.
        'summary' => $this->dataProcessorManager->process(
          $entity->get('body'),
        ),

        // Triggers processor in the article media field.
        'image' => $this->dataProcessorManager->process(
          $entity->get('field_media'),
          [
            'style' => 'article_card_400x300'
          ],
        ),
      ],
    ];
  }

}
```

And this processor can be trigger with

```php
$article = \Drupal::entityTypeManager()->getStorage('node')->load($aid);

$data = \Drupal::service('plugin.manager.flex_processor')->process(
  $article,
  [
    'variant' => 'card'
  ]
);

echo json_encode($data);
```
_output_
```json
// phpcs:disable
{
  "title" : "Title A",
  "type" : "article",
  "url" : "https://example.com/article/title-a",
  "fields" : {
    "summary" : "Bla bla bla bla...",
    "image" : "https://example.com/site/default/files/styles/article_card_400x300/public/2024-10/image-a.png",
  }
}
// phpcs:enable
```


### DataComponent processor example

Data Components are designed to process arrays structures. They were initially
created to reuse data wrappers structures, but they can have many other uses.

**Example of how to create a 'feed' processor plugin, to return a list of**
**processed items.**

First lets wrap the user data in a DataComponent. A DataComponent is a value
object with a bundle and  the contained data.

```php

// phpcs:ignore
use Drupal\flex_processor\Component\DataComponent;

$bundle = 'feed'

$articles = \Drupal::entityTypeManager()->getStorage('media')->loadMultiple([
  $aid1,
  $aid2,
]);

$feedData = ['items' => $articles];

$feedArticles = new DataComponent($bundle, $feedData);

```

And now lets create a plugin to process the feed data component.
The plugin type must be component, and the bundle the one specified in your data
component

> file: modules/custom/your_module/src/Plugin/ComponentDataProcessor/Feed.php
```php
namespace Drupal\your_module\Plugin\ComponentDataProcessor;

use Drupal\flex_processor\Plugin\ComponentDataProcessor;

/**
 * Returns the structured data of an entity.
 *
 * @DataProcessor(
 *   id = "feed",
 *   label = @Translation("Feed List"),
 *   type = "component",
 *   bundles = {
 *     "feed"
 *   }
 * )
 */
class Feed extends ComponentDataProcessor {

  /**
   * {@inheritdoc}
   */
  public function process(mixed $component, array $options = []): mixed {
    $items = [];

    // Process each feed item using the card variant.
    foreach ($component->get('items') as $item) {
      $items[] = $this->dataProcessorManager->process($item, [
        'variant' => 'card'
      ]);
    }

    // Return  the processed feed.
    return [
      'item' => $items,
      'total' => count($items),
    ];
  }

}
```

And this processor can be trigger with

```php

// phpcs:ignore

$data = \Drupal::service('plugin.manager.flex_processor')->process(
  $feedArticles
);

echo json_encode($data);
```
_output_

```json
// phpcs:disable
{
  "items" : [
    {
      "title" : "Title A",
      "type" : "article",
      "url" : "https://example.com/article/title-a",
      "fields" : {
        "summary" : "Bla bla bla bla...",
        "image" : "https://example.com/site/default/files/styles/article_card_400x300/public/2024-10/image-a.png",
      }
    },
    {
      "title" : "Title B",
      "type" : "article",
      "url" : "https://example.com/article/title-b",
      "fields" : {
        "summary" : "Bla bla bla bla...",
        "image" : "https://example.com/site/default/files/styles/article_card_400x300/public/2024-10/image-b.png",
      }
    }
  ],
  "total" : "2"
}
// phpcs:enable
```


### Processing Unknown data Structures

This module can identify by default 3 data types: Entities, Fields and
DataComponents; But it is simple to match any other Data structure with a
processor plugin, just create an even listener and check if the about to be
processed unknown object is the one you want

**Here is a sample of how to "register" a plugin for an unknown data object.**

Create the event listener to register a custom value object.

> file: modules/custom/your_module/src/EventSubscriber/DataComponentEventsSubscriber.php
```php

namespace Drupal\your_module\EventSubscriber;

use Drupal\flex_processor\Event\UnknownDataProcessorEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Subscribes to the UnknownDataProcessorEvent.
 */
class DataComponentEventsSubscriber implements EventSubscriberInterface {

  /**
   * {@inheritdoc}
   *
   * @return array
   *   The event names to listen for, and the methods that should be executed.
   */
  public static function getSubscribedEvents() {
    return [
      UnknownDataProcessorEvent::NAME => 'registerMyObject',
    ];
  }

  /**
   * Register processors for custom data structures.
   *
   * @param \Drupal\monks_data_processor\Event\UnknownDataProcessorEvent $event
   *   Config crud event.
   */
  public function registerMyObject(UnknownDataProcessorEvent $event) {

    // Get the data to be processed.
    $data = $event->getDataSource();

    // Verify data matches with your desired custom object.
    if (isset($data->pepe)) {

      // Set the plugin Type.
      $event->setDataSourceType('custom');

      // Set the plugin Bundle.
      $event->setDataSourceBundle('pepe');
    }
  }

}
```
Note this is only the listener, check this doc for a complete setup
[Drupal doc](https://www.drupal.org/docs/develop/creating-modules/subscribe-to-and-dispatch-events#s-my-first-drupal-8-event-subscriber)

Then create the plugin for the custom object, using the defined type and bundle.

> file: modules/custom/your_module/src/Plugin/Pepe.php
```php

namespace Drupal\your_module\Plugin;

use Drupal\flex_processor\Plugin\DataProcessorBase;

/**
 * Returns the structured data of an entity.
 *
 * @DataProcessor(
 *   id = "custom_pepe",
 *   label = @Translation("Custom Object!"),
 *   type = "custom",
 *   bundles = {
 *     "pepe"
 *   }
 * )
 */
class Pepe extends DataProcessorBase {

  /**
   * {@inheritdoc}
   */
  public function process(mixed $component, array $options = []): mixed {
    return [
      'type' => $this->getPluginDefinition()['id'],
      'pepe' => $component->pepe,
    ];
  }

}
```

And this processor can be trigger with

```php
$customObject = (object) ['pepe' => 'salta salta'];

$data = \Drupal::service('plugin.manager.flex_processor')->process(
  $customObject
);

echo json_encode($data);
```
_output_
```json
// phpcs:disable
{
  "type" : "custom_pepe",
  "pepe" : "salta salta"
}
// phpcs:enable
```
