<?php

namespace Drupal\flex_processor\Plugin;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Plugin\DefaultPluginManager;
use Drupal\flex_processor\Annotation\DataProcessor;
use Drupal\flex_processor\Component\DataComponentInterface;
use Drupal\flex_processor\Event\UnknownDataProcessorEvent;
use Drupal\flex_processor\Exception\MissingPluginImplementationException;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Provides the Task plugin plugin manager.
 */
class DataProcessorManager extends DefaultPluginManager {

  /**
   * Event Dispatcher Service.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $dispatcher;

  /**
   * Map of all defined plugin ids.
   *
   * @var array<string, string>
   */
  protected $pluginMap;

  public function __construct(
    \Traversable $namespaces,
    CacheBackendInterface $cache_backend,
    ModuleHandlerInterface $module_handler,
    EventDispatcherInterface $dispatcher,
  ) {
    parent::__construct(
      'Plugin',
      $namespaces,
      $module_handler,
      DataProcessorInterface::class,
      DataProcessor::class
    );
    $this->alterInfo('flex_processor_processor_info');
    $this->setCacheBackend($cache_backend, 'flex_processor_processor_plugins');
    $this->dispatcher = $dispatcher;
  }

  /**
   * Triggers the process of the given component.
   *
   * @param mixed $component
   *   The data component.
   * @param array<string, mixed> $options
   *   The options set to process the components.
   */
  public function process(mixed $component, array $options = []) : mixed {

    $defaultOptions = [
      'variant' => 'default',
      'langcode' => NULL,
    ];
    $options = array_merge($defaultOptions, $options);

    // Load the processor plugin that matches the component type.
    $componentType = $this->getComponentType($component, $options);
    $pluginId = $this->getProcessor($componentType, $options);

    /** @var \Drupal\flex_processor\Plugin\DataProcessorInterface $processor */
    $processor = $this->createInstance($pluginId);

    // Process components and retrieve the data.
    return $processor->preProcess($component, $options);
  }

  /**
   * Build the componentType of the given component.
   *
   * @param mixed $component
   *   The data component.
   * @param array<string, mixed> $options
   *   The options set to process the components.
   *
   * @return array<string, string>
   *   The component type definition.
   */
  protected function getComponentType(mixed $component, array $options = []) : array {
    $componentType = NULL;

    if ($component instanceof EntityInterface) {
      $componentType = [
        'type' => $component->getEntityTypeId(),
        'bundle' => $component->bundle(),
        'variant' => $options['variant'],
      ];
    }

    elseif ($component instanceof FieldItemListInterface) {
      $componentType = [
        'type' => 'field',
        'bundle' => $component->getFieldDefinition()->getType(),
        'variant' => $options['variant'],
      ];
    }

    elseif ($component instanceof DataComponentInterface) {
      $componentType = [
        'type' => 'component',
        'bundle' => $component->bundle(),
        'variant' => $options['variant'],
      ];
    }

    else {
      $event = new UnknownDataProcessorEvent($component, $options['variant']);
      $this->dispatcher->dispatch($event, UnknownDataProcessorEvent::NAME);
      $componentType = $event->getDataSourceComponentType();
    }

    return $componentType;
  }

  /**
   * Returns the processor ID for given entity.
   *
   * @param array<string, string> $componentType
   *   The data component definition.
   * @param array<string, mixed> $options
   *   The options set to process the components.
   */
  protected function getProcessor(array $componentType, array $options = []) : string {

    // Get a plugin that matches entity type, bundle and variant.
    $map = $this->getProcessorMap();

    $key = implode('.', [
      $componentType['type'],
      $componentType['bundle'],
      $componentType['variant'],
    ]);

    $keyDefault = implode('.', [
      $componentType['type'],
      $componentType['bundle'],
      'default',
    ]);

    // If no plugin is found for the specific variant, try the default variant.
    $processorId = $map[$key] ?? $map[$keyDefault] ?? '';

    if (empty($processorId)) {
      throw new MissingPluginImplementationException(sprintf("No DataProcessor plugin found for component of type '%s', bundle '%s' and variant '%s'", $componentType['type'], $componentType['bundle'], $componentType['variant']));
    }

    return $processorId;
  }

  /**
   * Loads a map of all plugin IDs keyed by type bundle and variant.
   *
   * @return array<string, string>
   *   Plugin ID map keyed by {$type}.{$bundle}.{variant}.
   */
  protected function getProcessorMap() : array {

    if (empty($this->pluginMap)) {
      $plugins = $this->getDefinitions();
      foreach ($plugins as $pluginId => $plugin) {
        foreach ($plugin['bundles'] as $bundle) {

          $key = implode('.', [
            $plugin['type'],
            $bundle,
            $plugin['variant'],
          ]);

          $this->pluginMap[$key] = $pluginId;

        }
      }
    }

    return $this->pluginMap;
  }

}
