<?php

namespace Drupal\flex_processor\Plugin\FieldDataProcessors;

use Drupal\Component\Utility\Xss;
use Drupal\flex_processor\Plugin\FieldDataProcessor;
use Drupal\image\Entity\ImageStyle;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Returns the (structured) data of a field.
 *
 * @DataProcessor(
 *   id = "field_image",
 *   label = @Translation("Image"),
 *   type = "field",
 *   bundles = {
 *     "image",
 *   }
 * )
 */
class FieldImage extends FieldDataProcessor {

  /**
   * Drupal\Core\Entity\EntityTypeManager definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * Drupal\Core\File\FileUrlGenerator service.
   *
   * @var \Drupal\Core\File\FileUrlGenerator
   */
  protected $urlGenerator;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $instance = parent::create(
      $container,
      $configuration,
      $plugin_id,
      $plugin_definition,
    );

    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->urlGenerator = $container->get('file_url_generator');

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function process(mixed $field, array $options = []): mixed {

    $files = $field->referencedEntities();
    $file = reset($files);
    if (empty($file)) {
      return NULL;
    }

    $uri = $file->uri->value;

    $url['original'] = $this->urlGenerator->generateAbsoluteString($uri);

    if (!empty($options['style'])) {
      $styles = (array) $options['style'];
      foreach ($styles as $style) {
        $url[$style] = $this->getStyledImageUrl($uri, $style);
      }
    }

    return [
      'url' => $url,
      'alt' => Xss::filter($field->alt),
      'meta' => [
        'width' => $field->width,
        'height' => $field->height,
        'orientation' => ($field->width >= $field->height) ? 'landscape' : 'portrait',
      ],
    ];

  }

  /**
   * Returns the URL of a styled image.
   *
   * @param string $uri
   *   The image file uri.
   * @param string $style
   *   The image style.
   *
   * @return string
   *   The URL of the styled image. and/Or the URL of the original image.
   *   This will generate the requested styled image.
   */
  protected function getStyledImageUrl(string $uri, string $style) : string {

    $style = ImageStyle::load($style);
    if ($style) {
      $url = $style->buildUrl($uri);
    }
    else {
      $url = $this->urlGenerator->generateAbsoluteString($uri);
    }

    return $url;
  }

}
